package fmcclient

import (
	"sync"
	"sync/atomic"
	"time"
)

type ringBufStatus struct {
	wSeq   int32
	rSeq   int32
	wWaits int32
	rWaits int32
}

func newRingBuf() *ringBuf {
	return &ringBuf{buf: make([][]byte, 8, 64)}
}

type ringBuf struct {
	ringBufStatus
	buf       [][]byte
	growMutex *sync.Mutex
}

func (r *ringBuf) Init() {
	r.growMutex = &sync.Mutex{}
	return
}

func (r *ringBuf) readNext() []byte {
	r.growMutex.Lock()
	if r.rSeq == atomic.LoadInt32(&r.wSeq) {
	BusyWait:
		for {
			atomic.AddInt32(&r.rWaits, 1)
			select {
			case <-time.After(100 * time.Microsecond):
				if r.rSeq < atomic.LoadInt32(&r.wSeq) {
					break BusyWait
				}
			}
		}
	}
	b := r.buf[r.rSeq&int32(len(r.buf)-1)]
	r.growMutex.Unlock()
	atomic.AddInt32(&r.rSeq, 1)
	return b
}

func (r *ringBuf) writeNextN(bufs [][]byte) {
	r.growMutex.Lock()
	bufSize := len(r.buf)
	lastWIndex := (r.wSeq + int32(len(bufs)) - 1) & int32(bufSize-1)
	if lastWIndex == atomic.LoadInt32(&r.rSeq)&int32(bufSize-1) {
	BusyWait:
		for {
			atomic.AddInt32(&r.wWaits, 1)
			select {
			case <-time.After(100 * time.Microsecond):
				if lastWIndex < atomic.LoadInt32(&r.rSeq)&int32(bufSize-1) {
					break BusyWait
				}
			}
		}
	}
	for i := 0; i < len(bufs); i++ {
		r.buf[(r.wSeq+int32(i))&int32(bufSize-1)] = bufs[i]
	}
	r.growMutex.Unlock()
	atomic.AddInt32(&r.wSeq, int32(len(bufs)))
	return
}

func (r *ringBuf) grow() error {
	return nil
}

func (r *ringBuf) status() ringBufStatus {
	return ringBufStatus{
		wSeq:   atomic.LoadInt32(&r.wSeq),
		rSeq:   atomic.LoadInt32(&r.rSeq),
		wWaits: atomic.LoadInt32(&r.wWaits),
		rWaits: atomic.LoadInt32(&r.rWaits),
	}
}
